from django.core.files.storage import default_storage
from rest_framework.renderers import BaseRenderer

from spillway import colors, styles
from spillway.compat import mapnik


class MapnikRenderer(BaseRenderer):
    """Renders Mapnik stylesheets to tiled PNG."""
    mapfile = default_storage.path('map.xml')
    media_type = 'image/png'
    format = 'png'
    charset = None
    render_style = 'binary'

    def __init__(self, *args, **kwargs):
        super(MapnikRenderer, self).__init__(*args, **kwargs)
        self.stylename = 'Spectral_r'
        self.context = {}
        m = mapnik.Map(256, 256)
        try:
            mapnik.load_map(m, str(self.mapfile))
        except RuntimeError:
            pass
        m.buffer_size = 128
        m.srs = '+init=epsg:3857'
        self.map = m

    def append_layer(self, object, stylename):
        try:
            style = self.map.find_style(stylename)
        except KeyError:
            style = styles.make_raster_style()
            self.map.append_style(stylename, style)
            rcolors = colors.colormap.get(stylename)
            bins = object.bin(k=len(rcolors))
            styles.add_colorizer_stops(style, bins, rcolors)
        try:
            layer = object.layer(self.context.get('band'))
        except AttributeError:
            pass
        else:
            layer.styles.append(stylename)
            # Must append layer to map *after* appending style to it.
            self.map.layers.append(layer)

    def render(self, object, accepted_media_type=None, renderer_context=None):
        self.context.update(renderer_context)
        response = renderer_context.get('response')
        if getattr(response, 'exception', False):
            return object
        img = mapnik.Image(self.map.width, self.map.height)
        bbox = renderer_context.get('bbox') if renderer_context else None
        stylename = str(renderer_context.get('style') or
                        getattr(object, 'style', self.stylename))
        self.append_layer(object, stylename)
        # Zero area bounding boxes are invalid.
        if bbox and bbox.area:
            bbox.transform(self.map.srs)
            self.map.zoom_to_box(mapnik.Box2d(*bbox.extent))
            mapnik.render(self.map, img)
        return img.tostring(self.format)


class MapnikJPEGRenderer(MapnikRenderer):
    """Renders Mapnik stylesheets to tiled JPEG."""
    media_type = 'image/jpeg'
    format = 'jpeg'



from django.http import StreamingHttpResponse
from rest_framework.generics import ListAPIView, ListCreateAPIView, RetrieveAPIView
from rest_framework.settings import api_settings

from spillway import filters, forms, mixins, pagination, renderers, serializers

_default_filters = tuple(api_settings.DEFAULT_FILTER_BACKENDS)
_default_renderers = tuple(api_settings.DEFAULT_RENDERER_CLASSES)


class BaseGeoView(mixins.ModelSerializerMixin):
    """Base view for models with geometry fields."""
    model_serializer_class = serializers.FeatureSerializer
    pagination_class = pagination.FeaturePagination
    filter_backends = _default_filters + (
        filters.SpatialLookupFilter, filters.GeoQuerySetFilter)
    renderer_classes = _default_renderers + (
        renderers.GeoJSONRenderer, renderers.KMLRenderer, renderers.KMZRenderer)


class GeoDetailView(BaseGeoView, RetrieveAPIView):
    """Generic detail view providing vector geometry representations."""


class GeoListView(BaseGeoView, ListAPIView):
    """Generic view for listing a geoqueryset."""


class GeoListCreateAPIView(BaseGeoView, ListCreateAPIView):
    """Generic view for listing or creating geomodel instances."""


class BaseRasterView(mixins.ModelSerializerMixin, mixins.QueryFormMixin):
    """Base view for raster models."""
    model_serializer_class = serializers.RasterModelSerializer
    query_form_class = forms.RasterQueryForm
    filter_backends = _default_filters
    renderer_classes = _default_renderers + (
        renderers.GeoTIFFZipRenderer,
        renderers.HFAZipRenderer,
    )

    def finalize_response(self, request, response, *args, **kwargs):
        response = super(BaseRasterView, self).finalize_response(
            request, response, *args, **kwargs)
        # Use streaming responses for GDAL formats.
        if isinstance(response.accepted_renderer,
                      renderers.gdal.BaseGDALRenderer):
            headers = response._headers
            response = StreamingHttpResponse(response.rendered_content)
            response._headers = headers
        return response

    def get_serializer_context(self):
        context = super(BaseRasterView, self).get_serializer_context()
        renderer = self.request.accepted_renderer
        context.update(format=renderer.format, **self.clean_params())
        return context


class RasterDetailView(BaseRasterView, RetrieveAPIView):
    """View providing access to a Raster model instance."""
    renderer_classes = _default_renderers + (
        renderers.GeoTIFFRenderer,
        renderers.HFARenderer,
    )


class RasterListView(BaseRasterView, ListAPIView):
    """View providing access to a Raster model QuerySet."""
    filter_backends = _default_filters + (filters.SpatialLookupFilter,)



from django.contrib.gis.db import models
from rest_framework import renderers, serializers
from greenwich.srs import SpatialReference
import numpy as np

from spillway import query, collections as sc
from spillway.fields import GeometryField, GDALField, NDArrayField

serializers.ModelSerializer.serializer_field_mapping.update({
    models.GeometryField: GeometryField,
    models.PointField: GeometryField,
    models.LineStringField: GeometryField,
    models.PolygonField: GeometryField,
    models.MultiPointField: GeometryField,
    models.MultiLineStringField: GeometryField,
    models.MultiPolygonField: GeometryField,
    models.GeometryCollectionField: GeometryField
})


class GeoModelSerializer(serializers.ModelSerializer):
    """Serializer class for GeoModels."""

    def __new__(cls, *args, **kwargs):
        cls.Meta.geom_field = getattr(cls.Meta, 'geom_field', None)
        return super(GeoModelSerializer, cls).__new__(cls, *args, **kwargs)

    def get_fields(self):
        """Returns a fields dict for this serializer with a 'geometry' field
        added.
        """
        fields = super(GeoModelSerializer, self).get_fields()
        # Set the geometry field name when it's undeclared.
        if not self.Meta.geom_field:
            for name, field in fields.items():
                if isinstance(field, GeometryField):
                    self.Meta.geom_field = name
                    break
        # Alter geometry field source based on requested format.
        try:
            renderer = self.context['request'].accepted_renderer
        except (AttributeError, KeyError):
            pass
        else:
            geom_field = fields.get(self.Meta.geom_field)
            obj = self.instance
            if self._is_paginated():
                obj = self.context['view'].queryset
            if geom_field and hasattr(obj, renderer.format):
                geom_field.source = renderer.format
        return fields

    def _is_paginated(self):
        try:
            return hasattr(self.context['view'].paginator, 'page')
        except (AttributeError, KeyError):
            return False


class FeatureListSerializer(serializers.ListSerializer):
    """Feature list serializer for GeoModels."""

    @property
    def data(self):
        return super(serializers.ListSerializer, self).data

    def to_representation(self, data):
        data = map(self.child.to_representation, data)
        try:
            srid = query.get_srid(self.instance)
        except AttributeError:
            srid = None
        return sc.FeatureCollection(features=data, crs=srid)


class FeatureSerializer(GeoModelSerializer):
    """Feature serializer for GeoModels."""

    def __new__(cls, *args, **kwargs):
        cls.Meta.list_serializer_class = getattr(
            cls.Meta, 'list_serializer_class', FeatureListSerializer)
        return super(FeatureSerializer, cls).__new__(cls, *args, **kwargs)

    @property
    def data(self):
        if not hasattr(self, '_data'):
            self._data = super(FeatureSerializer, self).data
            if 'crs' not in self._data:
                geom = getattr(self.instance, self.Meta.geom_field, None)
                if geom and geom.srid:
                    self._data['crs'] = sc.NamedCRS(geom.srid)
        return self._data

    def to_representation(self, instance):
        native = super(FeatureSerializer, self).to_representation(instance)
        geometry = native.pop(self.Meta.geom_field)
        pk = native.pop(instance._meta.pk.name, None)
        return sc.Feature(pk, geometry, native)

    def to_internal_value(self, data):
        if sc.has_features(data):
            for feat in data['features']:
                return self.to_internal_value(feat)
        try:
            sref = SpatialReference(data['crs']['properties']['name'])
        except KeyError:
            sref = None
        if not self.fields and self.Meta.geom_field is None:
            raise Exception('No geometry field found')
        record = {self.Meta.geom_field: data.get('geometry')}
        record.update(data.get('properties', {}))
        feature = super(FeatureSerializer, self).to_internal_value(record)
        if feature and sref:
            geom = feature[self.Meta.geom_field]
            geom.srid = sref.srid
        return feature


class RasterListSerializer(serializers.ListSerializer):
    """Raster list serializer for raster models."""

    def to_representation(self, data):
        data = super(RasterListSerializer, self).to_representation(data)
        periods = self.context.get('periods')
        attr = self.child.Meta.raster_field
        if periods and isinstance(self.child.fields[attr], NDArrayField):
            record = data[0]
            fill = getattr(record[attr], 'fill_value', None)
            arr = np.ma.array([row[attr] for row in data],
                              fill_value=fill, copy=False)
            try:
                arr = arr.reshape((periods, -1)).mean(axis=1)
            except ValueError:
                pass
            else:
                record[attr] = arr
                return [record]
        return data


class RasterModelSerializer(GeoModelSerializer):
    def __new__(cls, *args, **kwargs):
        cls.Meta.raster_field = getattr(cls.Meta, 'raster_field', None)
        cls.Meta.list_serializer_class = getattr(
            cls.Meta, 'list_serializer_class', RasterListSerializer)
        return super(RasterModelSerializer, cls).__new__(cls, *args, **kwargs)

    def get_fields(self):
        fields = super(RasterModelSerializer, self).get_fields()
        if not self.Meta.raster_field:
            for name, field in fields.items():
                if isinstance(field, serializers.FileField):
                    self.Meta.raster_field = name
                    break
        fieldname = self.Meta.raster_field
        request = self.context.get('request')
        renderer = getattr(request, 'accepted_renderer', None)
        if isinstance(renderer, renderers.JSONRenderer):
            if self.context.get('g'):
                fields[fieldname] = NDArrayField()
        elif isinstance(renderer, (renderers.BrowsableAPIRenderer,
                                   renderers.TemplateHTMLRenderer)):
            pass
        elif fieldname:
            fields['path'] = serializers.CharField(
                source='%s.path' % fieldname)
            fields['file'] = GDALField(source=fieldname)
        return fields



from django.core.exceptions import ImproperlyConfigured
from rest_framework.utils import encoders

try:
    # Django 1.8+
    from django.contrib.gis.db.models.lookups import gis_lookups as ALL_TERMS
except ImportError:
    # Django<=1.7
    from django.contrib.gis.db.models.sql.query import ALL_TERMS

try:
    import simplejson as json
except ImportError:
    import json
    JSONEncoder = encoders.JSONEncoder
else:
    # Workaround to support simplejson 2.2+, see
    # https://github.com/simplejson/simplejson/issues/37
    class JSONEncoder(json.JSONEncoder):
        default = encoders.JSONEncoder().default

try:
    import mapnik
except ImportError:
    class Mapnik(object):
        def __getattr__(self, attr):
            raise ImproperlyConfigured('Mapnik must be installed')

    mapnik = Mapnik()



import os
import shutil
import zipfile
import tempfile

from django.contrib.gis import forms
from django.contrib.gis import gdal
from django.contrib.gis.gdal.srs import SpatialReference, SRSException
from django.utils.translation import ugettext_lazy as _
from greenwich.geometry import Envelope

from spillway.compat import json


class CommaSepFloatField(forms.FloatField):
    """A Field for parsing a comma separated list of numeric values."""
    default_error_messages = {
        'invalid': _(u'Enter a comma separated list of numbers.'),
        'max_value': _(u'Ensure each value is less than or equal to %(limit_value)s.'),
        'min_value': _(u'Ensure each value is greater than or equal to %(limit_value)s.'),
    }

    def to_python(self, value):
        "Normalize data to a list of floats."
        if not value:
            return []
        return map(super(CommaSepFloatField, self).to_python, value.split(','))

    def run_validators(self, values):
        """Run validators for each item separately."""
        for val in values:
            super(CommaSepFloatField, self).run_validators(val)


class BoundingBoxField(CommaSepFloatField):
    """A Field for comma separated bounding box coordinates."""

    def __init__(self, srid=4326, *args, **kwargs):
        super(BoundingBoxField, self).__init__(*args, **kwargs)
        self.srid = srid

    def to_python(self, value):
        """Returns a GEOS Polygon from bounding box values."""
        value = super(BoundingBoxField, self).to_python(value)
        try:
            bbox = gdal.OGRGeometry.from_bbox(value).geos
        except (ValueError, AttributeError):
            return []
        bbox.srid = self.srid
        return bbox


class GeometryFileField(forms.FileField):
    """A Field for creating OGR geometries from file based sources."""

    def to_python(self, value):
        value = super(GeometryFileField, self).to_python(value)
        if value is None:
            return value
        filename = value.name
        tmpdir = None
        if zipfile.is_zipfile(value):
            tmpdir = tempfile.mkdtemp()
            zf = zipfile.ZipFile(value)
            # Extract all files from the temporary directory using only the
            # base file name, avoids security issues with relative paths in the
            # zip.
            for item in zf.namelist():
                filename = os.path.join(tmpdir, os.path.basename(item))
                with open(filename, 'wb') as f:
                    f.write(zf.read(item))
                if filename.endswith('.shp'):
                    break
        # Attempt to union all geometries from GDAL data source.
        try:
            geoms = gdal.DataSource(filename)[0].get_geoms()
            geom = reduce(lambda g1, g2: g1.union(g2), geoms)
            if not geom.srs:
                raise gdal.OGRException('Cannot determine SRS')
        except (gdal.OGRException, gdal.OGRIndexError):
            geom = None
        finally:
            if tmpdir and os.path.isdir(tmpdir):
                shutil.rmtree(tmpdir)
        return geom


class OGRGeometryField(forms.GeometryField):
    """A Field for creating OGR geometries."""

    def to_python(self, value):
        if value in self.empty_values:
            return None
        # Work with a single GeoJSON geometry or a Feature. Avoid parsing
        # overhead unless we have a true "Feature".
        if '"Feature",' in value:
            d = json.loads(value)
            value = json.dumps(d.get('geometry'))
        # Handle a comma delimited extent.
        elif list(value).count(',') == 3:
            value = Envelope(value.split(',')).polygon.ExportToWkt()
        try:
            geom = gdal.OGRGeometry(value)
        except (gdal.OGRException, TypeError, ValueError):
            raise forms.ValidationError(self.error_messages['invalid_geom'])
        if not geom.srs:
            geom.srid = self.srid or self.widget.map_srid
        return geom


class SpatialReferenceField(forms.IntegerField):
    """A Field for creating spatial reference objects."""

    def to_python(self, value):
        value = super(SpatialReferenceField, self).to_python(value)
        try:
            return SpatialReference(value)
        except (SRSException, TypeError):
            return None



from rest_framework import exceptions
from rest_framework.filters import BaseFilterBackend

from spillway import forms


class FormFilterBackend(BaseFilterBackend):
    queryset_form = None

    def filter_queryset(self, request, queryset, view):
        params = dict(request.query_params.dict(),
                      format=request.accepted_renderer.format,
                      **getattr(view, 'kwargs', {}))
        form = self.queryset_form(params, queryset)
        try:
            form.query()
        except ValueError:
            raise exceptions.ParseError(form.errors)
        return form.queryset


class GeoQuerySetFilter(FormFilterBackend):
    """A Filter for calling GeoQuerySet methods."""
    queryset_form = forms.GeometryQueryForm


class SpatialLookupFilter(FormFilterBackend):
    """A Filter providing backend supported spatial lookups like intersects,
    overlaps, etc.
    """
    queryset_form = forms.SpatialQueryForm


class TileFilter(FormFilterBackend):
    queryset_form = forms.MapTile
